#include "scene_descriptor.h"

#define ISQRT2 0.707106781186547
#define ISQRT3 0.577350269189626

SceneDescriptor::SceneDescriptor(Configuration *conf)
	: conf(conf)
{
	nBoxes = conf->nBoxes;
	boxes = new Box[nBoxes];
	baseIndices = new int[nBoxes];
	std::fill(baseIndices, baseIndices+nBoxes, 0);

	vertices = 0;
	for (int i = 0; i < nBoxes; i++)
	{
		boxes[i] = conf->boxes[i];
		boxes[i].allocate();
		Box &b = boxes[i];
		std::fill(b.acc, b.acc+(b.res_x*b.res_y), 0.0001f);
		baseIndices[i] = vertices;
		vertices += 6 * (b.res_x + 1) * (b.res_y + 1);
	}
	vertArray = new snow_vertex[vertices];
	for (int i = 0; i < vertices; i++)
	{
		vertArray[i].r = 1.0;
		vertArray[i].g = 1.0;
		vertArray[i].b = 1.0;
	}
	initVertexArray();
}

SceneDescriptor::~SceneDescriptor()
{
	delete [] vertArray;

	for (int i = 0; i < nBoxes; i++)
	{
		boxes[i].deallocate();
	}

	delete [] boxes;
	delete [] baseIndices;
}

float r = 0.199, g = 0.113, b = 0.051;

inline void
SceneDescriptor::calcColor(struct snow_vertex& v, float h)
{
	v.r = (0.6 - r) * min(h / conf->snowInvDepth, 1.0f) + r;
	v.g = (0.6 - g) * min(h / conf->snowInvDepth, 1.0f) + g;
	v.b = (0.65 - b) * min(h / conf->snowInvDepth, 1.0f) + b;
}

void
SceneDescriptor::initVertexArray()
{
	int baseIndex = 0;
	int index;

	for (int ib = 0; ib < nBoxes; ib++)
	{
		Box &b = boxes[ib];
		
		// compute vertex normals and vertex coordinates
		Vector3D n, n1, n2;
		Vector3D grad1, grad2;

		int w = b.res_x + 1;
		int h = b.res_y + 1;
		int i, j;


		// vertices at (x,y)
		for (int y = 0; y < h; y++)
		{
			j = y - 1;
			for (int x = 0; x < w; x++)
			{
				i = x - 1;
				index = 6 * (y * w + x) + baseIndex;

				if (x > 0 && x < w-1 && y > 0 && y < h-1) // inside domain
				{
					vertArray[index].x = b.pos.x + i * b.size.x / (b.res_x-1);
					vertArray[index].y = b.pos.y + j * b.size.y / (b.res_y-1);
	
					vertArray[index + 1].x = b.pos.x + (i+1) * b.size.x / (b.res_x-1);
					vertArray[index + 1].y = b.pos.y + j * b.size.y / (b.res_y-1);
					
					vertArray[index + 2].x = b.pos.x + (i+1) * b.size.x / (b.res_x-1);
					vertArray[index + 2].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);
	
					vertArray[index + 3].x = vertArray[index].x;
					vertArray[index + 3].y = vertArray[index].y;
	
					vertArray[index + 4].x = vertArray[index + 2].x;
					vertArray[index + 4].y = vertArray[index + 2].y;
				
					vertArray[index + 5].x = b.pos.x + i * b.size.x / (b.res_x-1);
					vertArray[index + 5].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);

				}
				else if (x == 0 && y > 0 && y < h-1) // left side
				{
					vertArray[index].x = b.pos.x;
					vertArray[index].y = b.pos.y + j * b.size.y / (b.res_y-1);
					vertArray[index].z = b.pos.z + b.size.z;
	
					vertArray[index + 1].x = b.pos.x;
					vertArray[index + 1].y = vertArray[index].y;
					
					vertArray[index + 2].x = b.pos.x;
					vertArray[index + 2].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);
	
					vertArray[index + 3].x = vertArray[index].x;
					vertArray[index + 3].y = vertArray[index].y;
					vertArray[index + 3].z = vertArray[index].z;
	
					vertArray[index + 4].x = vertArray[index + 2].x;
					vertArray[index + 4].y = vertArray[index + 2].y;
				
					vertArray[index + 5].x = b.pos.x;
					vertArray[index + 5].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);
					vertArray[index + 5].z = b.pos.z + b.size.z;
				}
				else if (x == w-1 && y > 0 && y < h-1) // right
				{
					vertArray[index].x = b.pos.x + b.size.x;
					vertArray[index].y = b.pos.y + j * b.size.y / (b.res_y-1);
	
					vertArray[index + 1].x = vertArray[index].x;
					vertArray[index + 1].y = vertArray[index].y;
					vertArray[index + 1].z = b.pos.z + b.size.z;
					
					vertArray[index + 2].x = vertArray[index].x;
					vertArray[index + 2].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);
					vertArray[index + 2].z = b.pos.z + b.size.z;
	
					vertArray[index + 3].x = vertArray[index].x;
					vertArray[index + 3].y = vertArray[index].y;
	
					vertArray[index + 4].x = vertArray[index + 2].x;
					vertArray[index + 4].y = vertArray[index + 2].y;
					vertArray[index + 4].z = vertArray[index + 2].z;
				
					vertArray[index + 5].x = vertArray[index].x;
					vertArray[index + 5].y = b.pos.y + (j+1) * b.size.y / (b.res_y-1);

				}
				else if (y == 0 && x > 0 && x < w-1) // bottom
				{
					vertArray[index].x = b.pos.x + i * b.size.x / (b.res_x-1);
					vertArray[index].y = b.pos.y;
					vertArray[index].z = b.pos.z + b.size.z;
	
					vertArray[index + 1].x = b.pos.x + (i+1) * b.size.x / (b.res_x-1);
					vertArray[index + 1].y = vertArray[index].y;
					vertArray[index + 1].z = b.pos.z + b.size.z;
					
					vertArray[index + 2].x = vertArray[index + 1].x;
					vertArray[index + 2].y = vertArray[index + 1].y;
	
					vertArray[index + 3].x = vertArray[index].x;
					vertArray[index + 3].y = vertArray[index].y;
					vertArray[index + 3].z = vertArray[index].z;
	
					vertArray[index + 4].x = vertArray[index + 2].x;
					vertArray[index + 4].y = vertArray[index + 2].y;
				
					vertArray[index + 5].x = vertArray[index].x;
					vertArray[index + 5].y = vertArray[index].y;
				}
				else if (y == h-1 && x > 0 && x < w-1) // top
				{
					vertArray[index].x = b.pos.x + i * b.size.x / (b.res_x-1);
					vertArray[index].y = b.pos.y + b.size.y;
	
					vertArray[index + 1].x = b.pos.x + (i+1) * b.size.x / (b.res_x-1);
					vertArray[index + 1].y = b.pos.y + b.size.y;
					
					vertArray[index + 2].x = vertArray[index + 1].x;
					vertArray[index + 2].y = b.pos.y + b.size.y;
					vertArray[index + 2].z = b.pos.z + b.size.z;
	
					vertArray[index + 3].x = vertArray[index].x;
					vertArray[index + 3].y = vertArray[index].y;
					vertArray[index + 3].z = vertArray[index].z;
	
					vertArray[index + 4].x = vertArray[index + 2].x;
					vertArray[index + 4].y = vertArray[index + 2].y;
					vertArray[index + 4].z = vertArray[index + 2].z;
				
					vertArray[index + 5].x = b.pos.x + i * b.size.x / (b.res_x-1);
					vertArray[index + 5].y = b.pos.y + b.size.y;
					vertArray[index + 5].z = b.pos.z + b.size.z;

				}

				// NORMAL
				if (x == 1 && y == 1) // bottom left corner
				{
					vertArray[index - 6].nx = -ISQRT3; // 1/sqrt(2)
					vertArray[index - 6].ny = -ISQRT3;
					vertArray[index - 6].nz = ISQRT3;
					vertArray[index - 3].nx = -ISQRT3;
					vertArray[index - 3].ny = -ISQRT3;
					vertArray[index - 3].nz = ISQRT3;
					vertArray[index - w*6].nx = -ISQRT3;
					vertArray[index - w*6].ny = -ISQRT3;
					vertArray[index - w*6].nz = ISQRT3;
					vertArray[index - w*6 + 3].nx = -ISQRT3;
					vertArray[index - w*6 + 3].ny = -ISQRT3;
					vertArray[index - w*6 + 3].nz = ISQRT3;
					calcColor(vertArray[index-6], b.acc[j * b.res_x + i]);
					calcColor(vertArray[index-3], b.acc[j * b.res_x + i]);
					calcColor(vertArray[index-w*6], b.acc[j * b.res_x + i]);
					calcColor(vertArray[index-w*6+3], b.acc[j * b.res_x + i]);

					grad1 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * (-ISQRT3 + n1.x);
					vertArray[index].ny = 0.5 * (-ISQRT3 + n1.y);
					vertArray[index].nz = 0.5 * (ISQRT3 + n1.z);
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
				}
				else if (x == 1 && y == h-1) // top left corner
				{
					vertArray[index - 6].nx = -ISQRT3;
					vertArray[index - 6].ny = ISQRT3;
					vertArray[index - 6].nz = ISQRT3;
					vertArray[index + 5].nx = -ISQRT3;
					vertArray[index + 5].ny = ISQRT3;
					vertArray[index + 5].nz = ISQRT3;

					grad2 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad1 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * (-ISQRT3 + n1.x);
					vertArray[index].ny = 0.5 * (ISQRT3 + n1.y);
					vertArray[index].nz = 0.5 * (ISQRT3 + n1.z);
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;

				}
				else if (x == w-1 && y == 1) // right bottom corner
				{
					vertArray[index + 1].nx = ISQRT3;
					vertArray[index + 1].ny = -ISQRT3;
					vertArray[index + 1].nz = ISQRT3;

					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * j + i-1] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * (ISQRT3 + n1.x);
					vertArray[index].ny = 0.5 * (-ISQRT3 + n1.y);
					vertArray[index].nz = 0.5 * (ISQRT3 + n1.z);
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index -  5].nx = vertArray[index].nx;
					vertArray[index -  5].ny = vertArray[index].ny;
					vertArray[index -  5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
				}
				else if (x == w-1 && y == h-1)
				{
					vertArray[index - w*6 + 2].nx = ISQRT3;
					vertArray[index - w*6 + 2].ny = ISQRT3;
					vertArray[index - w*6 + 2].nz = ISQRT3;
					vertArray[index - w*6 + 4].nx = ISQRT3;
					vertArray[index - w*6 + 4].ny = ISQRT3;
					vertArray[index - w*6 + 4].nz = ISQRT3;
					vertArray[index - 4].nx = ISQRT3;
					vertArray[index - 4].ny = ISQRT3;
					vertArray[index - 4].nz = ISQRT3;
					vertArray[index - 2].nx = ISQRT3;
					vertArray[index - 2].ny = ISQRT3;
					vertArray[index - 2].nz = ISQRT3;

					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * j + i-1] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * (ISQRT3 + n1.x);
					vertArray[index].ny = 0.5 * (-ISQRT3 + n1.y);
					vertArray[index].nz = 0.5 * (ISQRT3 + n1.z);
					vertArray[index -  5].nx = vertArray[index].nx;
					vertArray[index -  5].ny = vertArray[index].ny;
					vertArray[index -  5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
				}
				else if (x == 0 && y > 1 && y < h-1) // left border
				{
					vertArray[index].nx = -ISQRT2; //// HER!
					vertArray[index].ny = 0.0;
					vertArray[index].nz = ISQRT2;
					vertArray[index + 3].nx = -ISQRT2;
					vertArray[index + 3].ny = 0.0;
					vertArray[index + 3].nz = ISQRT2;
					vertArray[index - 6 * w + 5].nx = -ISQRT2;
					vertArray[index - 6 * w + 5].ny = 0.0;
					vertArray[index - 6 * w + 5].nz = ISQRT2;
				}
				else if (x == w-1 && y > 1 && y < h-1)
				{
					// x == w-1
					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);

					vertArray[index].nx = 0.5 * n1.x + 0.5 * ISQRT2;
					vertArray[index].ny = 0.5 * n1.y;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT2;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;

					// x == w
					vertArray[index + 1].nx = ISQRT2;
					vertArray[index + 1].ny = 0.0;
					vertArray[index + 1].nz = ISQRT2;
					vertArray[index - w*6 + 2].nx = ISQRT2;
					vertArray[index - w*6 + 2].ny = 0.0;
					vertArray[index - w*6 + 2].nz = ISQRT2;
					vertArray[index - w*6 + 4].nx = ISQRT2;
					vertArray[index - w*6 + 4].ny = 0.0;
					vertArray[index - w*6 + 4].nz = ISQRT2;
				}
				else if (y == 0 && x > 1 && x < w-1)
				{
					vertArray[index].nx = 0.0;
					vertArray[index].ny = -ISQRT2;
					vertArray[index].nz = ISQRT2;
					vertArray[index + 3].nx = 0.0;
					vertArray[index + 3].ny = -ISQRT2;
					vertArray[index + 3].nz = ISQRT2;
					vertArray[index - 5].nx = 0.0;
					vertArray[index - 5].ny = -ISQRT2;
					vertArray[index - 5].nz = ISQRT2;
				}
				else if (y == h-1 && x > 1 && x < w-1)
				{
					// y == h-1
					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);

					vertArray[index].nx = 0.5 * n1.x;
					vertArray[index].ny = 0.5 * n1.y + 0.5 * ISQRT2;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT2;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;

					// y == h
					vertArray[index + 5].nx = 0.0;
					vertArray[index + 5].ny = ISQRT2;
					vertArray[index + 5].nz = ISQRT2;
					vertArray[index - 4].nx = 0.0;
				   	vertArray[index - 4].ny = ISQRT2;
					vertArray[index - 4].nz = ISQRT2;
					vertArray[index - 2].nx = 0.0;
					vertArray[index - 2].ny = ISQRT2;
					vertArray[index - 2].nz = ISQRT2;
				}
				else if (x == 1 && y > 1 && y < h-1)
				{
					grad1 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * n1.x - 0.5 * ISQRT2;
					vertArray[index].ny = 0.5 * n1.y;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT2;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
				}
				else if (y == 1 && x > 1 && x < w-1)
				{
					
					grad1 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * n1.x;
					vertArray[index].ny = 0.5 * n1.y - 0.5 * ISQRT2;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT2;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
				}
				else if (x > 1 && y > 1 && x < w-1 && y < h-1)
				{	
					grad1 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					n2 = grad1.cross(grad2);
					n = n1 * 0.5f + n2 * 0.5f;
					n.normalize();

					//cout << n.x << ", " << n.y << ", " << n.z << endl;
	
					vertArray[index].nx = n.x;
					vertArray[index].ny = n.y;
					vertArray[index].nz = n.z;
					
					vertArray[index + 3].nx = n.x;
					vertArray[index + 3].ny = n.y;
					vertArray[index + 3].nz = n.z;
					
					vertArray[index - 5].nx = n.x;
					vertArray[index - 5].ny = n.y;
					vertArray[index - 5].nz = n.z;
	
					vertArray[index - (w + 1) * 6 + 2].nx = n.x;
					vertArray[index - (w + 1) * 6 + 2].ny = n.y;
					vertArray[index - (w + 1) * 6 + 2].nz = n.z;
	
					vertArray[index - (w + 1) * 6 + 4].nx = n.x;
					vertArray[index - (w + 1) * 6 + 4].ny = n.y;
					vertArray[index - (w + 1) * 6 + 4].nz = n.z;
	
					vertArray[index - w * 6 + 5].nx = n.x;
					vertArray[index - w * 6 + 5].ny = n.y;
					vertArray[index - w * 6 + 5].nz = n.z;
				}
			}
		}
		baseIndex += 6 * (b.res_x + 1) * (b.res_y + 1);
	}

}

void
SceneDescriptor::genVertexArray(SnowcoverJob& job)
{
	for (int ib = 0; ib < nBoxes; ib++)
	{
		Box &b = boxes[ib];
		
		// compute vertex normals and vertex coordinates
		Vector3D n, n1, n2;
		Vector3D grad1, grad2;

		int w = b.res_x + 1;
		int h = b.res_y + 1;
		int i, j;


		// vertices at (x,y)
		for (int y = job.yNormalsFrom[ib]; y < job.yNormalsTo[ib]; y++)
		{
			j = y - 1;
			for (int x = 2; x < w-1; x++)
			{
				i = x - 1;
				//
				//if (x > 1 && y > 1 && x < w-1 && y < h-1)

				int index = 6 * (y * w + x) + baseIndices[ib];
					grad1.x = 1.0f;
					grad1.y = 0.0f;
					grad1.z = b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i];
					grad2.x = 0.0f;
					grad2.y = 1.0f;
					grad2.z = b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i];
					n1.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n1.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n1.z = grad1.x * grad2.y - grad1.y * grad2.x;
					grad1.x = -1.0f;
					grad1.z = b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i];
					grad2.y = -1.0f;
					grad2.z = b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i];
					n2.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n2.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n2.z = grad1.x * grad2.y - grad1.y * grad2.x;
					n.x = 0.5f * (n1.x + n2.x);
					n.y = 0.5f * (n1.y + n2.y);
					n.z = 0.5f * (n1.z + n2.z);
					float ilength = 1.0f /(sqrtf(n.x*n.x + n.y*n.y + n.z*n.z));
					n.x *= ilength;
					n.y *= ilength;
					n.z *= ilength;

					//cout << n.x << ", " << n.y << ", " << n.z << endl;
	
					vertArray[index].nx = n.x;
					vertArray[index].ny = n.y;
					vertArray[index].nz = n.z;
					
					vertArray[index + 3].nx = n.x;
					vertArray[index + 3].ny = n.y;
					vertArray[index + 3].nz = n.z;
					
					vertArray[index - 5].nx = n.x;
					vertArray[index - 5].ny = n.y;
					vertArray[index - 5].nz = n.z;
	
					vertArray[index - (w + 1) * 6 + 2].nx = n.x;
					vertArray[index - (w + 1) * 6 + 2].ny = n.y;
					vertArray[index - (w + 1) * 6 + 2].nz = n.z;
	
					vertArray[index - (w + 1) * 6 + 4].nx = n.x;
					vertArray[index - (w + 1) * 6 + 4].ny = n.y;
					vertArray[index - (w + 1) * 6 + 4].nz = n.z;
	
					vertArray[index - w * 6 + 5].nx = n.x;
					vertArray[index - w * 6 + 5].ny = n.y;
					vertArray[index - w * 6 + 5].nz = n.z;
			}
		}

		// vertices at (x,y)
		for (int y = job.yVerticesFrom[ib]; y < job.yVerticesTo[ib]; y++)
		{
			j = y - 1;
			for (int x = 1; x < w-1; x++)
			{
				i = x - 1;
				int index = 6 * (y * w + x) + baseIndices[ib];

				//if (x > 0 && x < w-1 && y > 0 && y < h-1) // inside domain
					vertArray[index].z = b.pos.z + b.size.z + b.acc[j * b.res_x + i];
	
					vertArray[index + 1].z = b.pos.z + b.size.z + b.acc[j * b.res_x + (i+1)];
					
					vertArray[index + 2].z = b.pos.z + b.size.z + b.acc[(j+1) * b.res_x + (i+1)];
	
					vertArray[index + 3].z = vertArray[index].z;
	
					vertArray[index + 4].z = vertArray[index + 2].z;
				
					vertArray[index + 5].z = b.pos.z + b.size.z + b.acc[(j+1) * b.res_x + i];

					calcColor(vertArray[index], b.acc[j * b.res_x + i]);
					calcColor(vertArray[index+1], b.acc[j * b.res_x + i+1]);
					calcColor(vertArray[index+2], b.acc[(j+1) * b.res_x + i+1]);
					calcColor(vertArray[index+3], b.acc[j * b.res_x + i]);
					calcColor(vertArray[index+4], b.acc[(j+1) * b.res_x + i+1]);
					calcColor(vertArray[index+5], b.acc[(j+1) * b.res_x + i]);
			}
		}


		//if (x == 0 && y > 0 && y < h-1) // left side
		for (int y = job.yVerticesFrom[ib]; y < job.yVerticesTo[ib]; y++)
		{
			j = y - 1;
			int index = 6 * (y * w) + baseIndices[ib];
			calcColor(vertArray[index], b.acc[j * b.res_x]);

			vertArray[index + 1].z = b.pos.z + b.size.z + b.acc[j * b.res_x];
			calcColor(vertArray[index + 1], b.acc[j * b.res_x]);
			
			vertArray[index + 2].z = b.pos.z + b.size.z + b.acc[(j+1) * b.res_x];
			calcColor(vertArray[index + 2], b.acc[(j+1) * b.res_x]);

			calcColor(vertArray[index + 3], b.acc[j * b.res_x]);

			vertArray[index + 4].z = vertArray[index + 2].z;
			calcColor(vertArray[index + 4], b.acc[(j+1) * b.res_x]);
				
			calcColor(vertArray[index + 5], b.acc[(j+1) * b.res_x]);
		}

		//if (x == w-1 && y > 0 && y < h-1) // right
		for (int y = job.yVerticesFrom[ib]; y < job.yVerticesTo[ib]; y++)
		{
			j = y - 1;
			i = w - 2;
			int index = 6 * (y * w + w-1) + baseIndices[ib];

			vertArray[index].z = b.pos.z + b.size.z + b.acc[j * b.res_x + i];
			calcColor(vertArray[index], b.acc[j * b.res_x + i]);

			calcColor(vertArray[index + 1], b.acc[j * b.res_x + i]);
				
			calcColor(vertArray[index + 2], b.acc[(j+1) * b.res_x + i]);
	
			vertArray[index + 3].z = vertArray[index].z;
			calcColor(vertArray[index + 3], b.acc[j * b.res_x + i]);
	
			calcColor(vertArray[index + 4], b.acc[(j+1) * b.res_x + i]);
				
			vertArray[index + 5].z = b.pos.z + b.size.z + b.acc[(j+1) * b.res_x + i];
			calcColor(vertArray[index + 5], b.acc[(j+1) * b.res_x + i]);
		}

		//if (y == 0 && x > 0 && x < w-1) // bottom
		for (int x = job.xVerticesFrom[ib]; x < job.xVerticesTo[ib]; x++)
		{
			i = x - 1;
			int index = 6 * x + baseIndices[ib];

			calcColor(vertArray[index], b.acc[i]);
	
			calcColor(vertArray[index + 1], b.acc[i+1]);
					
			vertArray[index + 2].z = b.pos.z + b.size.z + b.acc[i+1];
			calcColor(vertArray[index + 2], b.acc[i+1]);
	
			calcColor(vertArray[index + 3], b.acc[i]);
	
			vertArray[index + 4].z = vertArray[index + 2].z;
			calcColor(vertArray[index + 4], b.acc[i+1]);
				
			vertArray[index + 5].z = b.pos.z + b.size.z + b.acc[i];
			calcColor(vertArray[index + 5], b.acc[i]);
		}

		//if (y == h-1 && x > 0 && x < w-1) // top
		for (int x = job.xVerticesFrom[ib]; x < job.xVerticesTo[ib]; x++)
		{
			i = x - 1;
			j = h-2;
			int index = 6 * ((h-1) * w + x) + baseIndices[ib];

			vertArray[index].z = b.pos.z + b.size.z + b.acc[j * b.res_x + i];
			calcColor(vertArray[index], b.acc[j * b.res_x + i]);
	
			vertArray[index + 1].z = b.pos.z + b.size.z + b.acc[j * b.res_x + i + 1];
			calcColor(vertArray[index + 1], b.acc[j * b.res_x + i + 1]);
					
			calcColor(vertArray[index + 2], b.acc[j * b.res_x + i + 1]);
	
			vertArray[index + 3].z = vertArray[index].z;
			calcColor(vertArray[index + 3], b.acc[j * b.res_x + i]);
	
			calcColor(vertArray[index + 4], b.acc[j * b.res_x + i + 1]);
			
			calcColor(vertArray[index + 5], b.acc[j * b.res_x + i]);
		}

		//if (x == w-1 && y > 1 && y < h-1)
		for (int y = job.yNormalsFrom[ib]; y < job.yNormalsTo[ib]; y++)
		{
			j = y - 1;
			i = w-2;
			int index = 6 * (y * w + w-1) + baseIndices[ib];
					// x == w-1
					grad1.x = -1.0f;
					grad1.y = 0.0f;
					grad1.z = b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i];
					grad2.x = 0.0f;
					grad2.y = -1.0f;
					grad2.z = b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i];
					n1.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n1.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n1.z = grad1.x * grad2.y - grad1.y * grad2.x;

					vertArray[index].nx = 0.5 * n1.x + 0.5 * ISQRT3;
					vertArray[index].ny = 0.5 * n1.y;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT3;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
		}

		//if (y == h-1 && x > 1 && x < w-1)
		for (int x = job.xNormalsFrom[ib]; x < job.xNormalsTo[ib]; x++)
		{
			j = h-2;
			i = x - 1;
			int index = 6 * ((h-1) * w + x) + baseIndices[ib];
					// y == h-1
					grad1.x = -1.0f;
					grad1.y = 0.0f;
					grad1.z = b.acc[b.res_x * j + i - 1] - b.acc[b.res_x * j + i];
					grad2.x = 0.0f;
					grad2.y = -1.0f;
					grad2.z = b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i];
					n1.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n1.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n1.z = grad1.x * grad2.y - grad1.y * grad2.x;

					vertArray[index].nx = 0.5 * n1.x;
					vertArray[index].ny = 0.5 * n1.y + 0.5 * ISQRT3;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT3;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
		}

		//if (x == 1 && y > 1 && y < h-1)
		for (int y = job.yNormalsFrom[ib]; y < job.yNormalsTo[ib]; y++)
		{
			j = y - 1;
			i = 0;
				int index = 6 * (y * w + 1) + baseIndices[ib];
					grad1.x = 1.0f;
					grad1.y = 0.0f;
					grad1.z = b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i];
					grad2.x = 0.0f;
					grad2.y = 1.0f;
					grad2.z = b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i];
					n1.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n1.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n1.z = grad1.x * grad2.y - grad1.y * grad2.x;
					vertArray[index].nx = 0.5 * n1.x - 0.5 * ISQRT3;
					vertArray[index].ny = 0.5 * n1.y;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT3;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
		}
		//if (y == 1 && x > 1 && x < w-1)
		for (int x = job.xNormalsFrom[ib]; x < job.xNormalsTo[ib]; x++)
		{
			j = 0;
			i = x - 1;
			int index = 6 * (w + x) + baseIndices[ib];

					grad1.x = 1.0f;
					grad1.y = 0.0f;
					grad1.z = b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i];
					grad2.x = 0.0f;
					grad2.y = 1.0f;
					grad2.z = b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i];
					n1.x = grad1.y * grad2.z - grad1.z * grad2.y;
					n1.y = grad1.z * grad2.x - grad1.x * grad2.z;
					n1.z = grad1.x * grad2.y - grad1.y * grad2.x;
					vertArray[index].nx = 0.5 * n1.x;
					vertArray[index].ny = 0.5 * n1.y - 0.5 * ISQRT3;
					vertArray[index].nz = 0.5 * n1.z + 0.5 * ISQRT3;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 2].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 2].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 2].nz = vertArray[index].nz;
					vertArray[index - 6*(w+1) + 4].nx = vertArray[index].nx;
					vertArray[index - 6*(w+1) + 4].ny = vertArray[index].ny;
					vertArray[index - 6*(w+1) + 4].nz = vertArray[index].nz;
					vertArray[index - 6*w + 5].nx = vertArray[index].nx;
					vertArray[index - 6*w + 5].ny = vertArray[index].ny;
					vertArray[index - 6*w + 5].nz = vertArray[index].nz;
		}
		


				//if (x == 1 && y == 1) // bottom left corner
				//{
				//x = 1;
				//y = 1;
				i = 0;
				j = 0;
				int index = 6 * (w + 1) + baseIndices[ib];
					grad1 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * -ISQRT3 + 0.5 * n1.x;
					vertArray[index].ny = 0.5 * -ISQRT3 + 0.5 * n1.y;
					vertArray[index].nz = 0.5 * ISQRT3 + 0.5 * n1.z;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index - 5].nx = vertArray[index].nx;
					vertArray[index - 5].ny = vertArray[index].ny;
					vertArray[index - 5].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
				//}
				//else if (x == 1 && y == h-1) // top left corner
				//{
				//x = 1;
				//y = h-1;
				i = 0;
				j = h-2;
				index = 6 * ((h-1) * w + 1) + baseIndices[ib];
					grad2 = Vector3D(1.0f, 0.0f, b.acc[b.res_x * j + i + 1] - b.acc[b.res_x * j + i]);
					grad1 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * -ISQRT3 + 0.5 * n1.x;
					vertArray[index].ny = 0.5 * ISQRT3 + 0.5 * n1.y;
					vertArray[index].nz = 0.5 * ISQRT3 + 0.5 * n1.z;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;

				//}
				//else if (x == w-1 && y == 1) // right bottom corner
				//{
				//x = w-1;
				//y = 1;
				i = w-2;
				j = 0;
				index = 6 * (w + w-1) + baseIndices[ib];
					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * (j+1) + i] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, 1.0f, b.acc[b.res_x * j + i-1] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * ISQRT3 + 0.5 * n1.x;
					vertArray[index].ny = 0.5 * -ISQRT3 + 0.5 * n1.y;
					vertArray[index].nz = 0.5 * ISQRT3 + 0.5 * n1.z;
					vertArray[index + 3].nx = vertArray[index].nx;
					vertArray[index + 3].ny = vertArray[index].ny;
					vertArray[index + 3].nz = vertArray[index].nz;
					vertArray[index -  5].nx = vertArray[index].nx;
					vertArray[index -  5].ny = vertArray[index].ny;
					vertArray[index -  5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
				//}
				//else if (x == w-1 && y == h-1)
				//{
				//x = w-1;
				//y = h-1;
				i = w-2;
				j = h-2;
				index = 6 * ((h-1) * w + (w-1)) + baseIndices[ib];
					grad1 = Vector3D(-1.0f, 0.0f, b.acc[b.res_x * (j-1) + i] - b.acc[b.res_x * j + i]);
					grad2 = Vector3D(0.0f, -1.0f, b.acc[b.res_x * j + i-1] - b.acc[b.res_x * j + i]);
					n1 = grad1.cross(grad2);
					vertArray[index].nx = 0.5 * ISQRT3 + 0.5 * n1.x;
					vertArray[index].ny = 0.5 * -ISQRT3 + 0.5 * n1.y;
					vertArray[index].nz = 0.5 * ISQRT3 + 0.5 * n1.z;
					vertArray[index -  5].nx = vertArray[index].nx;
					vertArray[index -  5].ny = vertArray[index].ny;
					vertArray[index -  5].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 2].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 2].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 2].nz = vertArray[index].nz;
					vertArray[index - (w+1)*6 + 4].nx = vertArray[index].nx;
					vertArray[index - (w+1)*6 + 4].ny = vertArray[index].ny;
					vertArray[index - (w+1)*6 + 4].nz = vertArray[index].nz;
					vertArray[index - w*6 + 5].nx = vertArray[index].nx;
					vertArray[index - w*6 + 5].ny = vertArray[index].ny;
					vertArray[index - w*6 + 5].nz = vertArray[index].nz;
				
		
	}
}


