#include "snow_simulator.h"

#include "build_solution_vector_job.h"
#include "snowflake_job.h"
#include "advect_job.h"
#include "solve_poisson_job.h"
#include "snowcover_job.h"
#include "render_job.h"
#include "prepare_timestep_job.h"
#include "advance_snowflakes_job.h"

SnowSimulator::SnowSimulator(	Configuration *conf )
	: conf(conf)
{
	int i;

	// Initialize simulation
	scene = new SceneDescriptor(conf);
	windField = new WindField(conf, scene);
	fallingSnow = new FallingSnowSimulation(conf, scene, windField);
	renderer = new SnowRenderer(conf, scene, windField, fallingSnow);
	barrier = new Barrier(conf->threads);

	// create threads
	threads = new SimulationThread*[conf->threads-1];

	for (i = 0; i < conf->threads-1; i++)
	{
		threads[i] = new SimulationThread(i+1, this);
	}

	frames = 0;

	initJobs();
}

SnowSimulator::~SnowSimulator()
{
	int i;
	conf->state.interrupted = true;
	
	for (i = 0; i < conf->threads-1; i++)
	{
		threads[i]->join();
		delete threads[i];
	}

	delete [] threads;
	delete barrier;
	delete renderer;
	delete fallingSnow;
	delete windField;
	delete scene;

	for (i = 0; i < nJobs; i++)
		delete jobs[i];
	
	delete [] jobs;
}

void 
SnowSimulator::runSimulation()
{
	int i;

	timer.increment();
	
	for (i = 0; i < conf->threads-1; i++)
	{
		threads[i]->start();
	}

	nextPrintTime = timer.time() + conf->printFpsInterval;

	if (conf->parallelMode == PARALLEL_DATA)
		mainLoopDataParallel(0);
	else if (conf->parallelMode == PARALLEL_PIPELINE)
		mainLoopPipelined(0);
	else
		mainLoopSerial();
}

/**
 * Returns the next available task a threads.
 * Last task must be provided, to choose the next task in the best possible way
 */

Job *
SnowSimulator::getJob(int threadId, Job *lastJob)
{
	int i = 0;
	Job *job = NULL;

	// If thread is main thread, and rendering is ready, choose that
	if (jobs[nJobs-1]->state == JOB_STATE_READY && jobs[nJobs-1]->forcedThread == threadId)
	{
		job = jobs[nJobs-1];
		job->state = JOB_STATE_RUNNING;
	}
	else
	// If the preferred next task is ready, take that
	if (lastJob != NULL && lastJob->preferredSuccessor->state == JOB_STATE_READY)
	{
		queueMutex.lock();
		if ((lastJob->preferredSuccessor->state == JOB_STATE_READY) && (
			(lastJob->preferredSuccessor->forcedThread < 0) || 
			(lastJob->preferredSuccessor->forcedThread == threadId)))
		{
			job = lastJob->preferredSuccessor;
			job->state = JOB_STATE_RUNNING;
		}
		queueMutex.unlock();
	}
	
	// Traverse the queue, looking for ready tasks
	while (job == NULL && !conf->state.interrupted)
	{
		if (jobs[i]->state == JOB_STATE_READY && (
			jobs[i]->forcedThread < 0 || jobs[i]->forcedThread == threadId))
		{
			queueMutex.lock();
			if (jobs[i]->state == JOB_STATE_READY)
			{
				job = jobs[i];
				job->state = JOB_STATE_RUNNING;
			}
			queueMutex.unlock();
		}
		i++;
		i %= nJobs;
	}

	/*printMutex.lock();
	cout << "Thread " << threadId << " got job: " << job->desc << endl;
	job->print();
	printMutex.unlock();*/
	return job;
}

/**
 * Place a task back in the queue, and update children
 */
void
SnowSimulator::finishJob(Job *job)
{
	queueMutex.lock();
	job->state = JOB_STATE_BLOCKED;

	for (int i = 0; i < job->nChildren; i++)
	{
		job->children[i]->nParentsDone++;
		if (job->children[i]->nParentsDone == job->children[i]->nParents)
		{
			job->children[i]->state = JOB_STATE_READY;
			job->children[i]->nParentsDone = 0;
		}
	}
	queueMutex.unlock();
}

void
SnowSimulator::mainLoopPipelined(int threadId)
{
	Job *job = NULL;
	while (!conf->state.interrupted)
	{
		job = getJob(threadId, job);
		
		/*if (frames > 50 && frames < 55)
		{
			printMutex.lock();
			cout << "Thread " << threadId << " got job: " << job->desc <<
					". Waited for " << t.idelta() << "microseconds." << endl;
			job->print();
			printMutex.unlock();
		}*/
/*		if (frames == 50 && threadId == 0) {
	for (int i = 0; i < nJobs; i++)
	{
		Job *job = jobs[i];
		job->print();
		cout << "   Children: ------------" << endl;
		for (int j = 0; j < job->nChildren; j++)
		{
			job->children[j]->print();
		}
		cout << "   ----------------------" << endl;
	}
		}*/
		if (job != NULL)
			doJob(job);
		else
			return;
		finishJob(job);
	}
}

void 
SnowSimulator::mainLoopDataParallel(int threadId)
{
	while (!conf->state.interrupted)
	{
		// Self-advect velocities
		windField->advect(*(AdvectJob*)jobs[conf->threads*0+threadId], timer.ddelta());
		barrier->wait();

		// Create right-hand side of Ax = b
		windField->buildSolutionVector(*(BuildSolutionVectorJob*)jobs[conf->threads*1+threadId], timer.ddelta());
		barrier->wait();

		// Solve Poisson equation using SOR
		for (int i = 0; i < conf->sorIter; i++)
		{
			windField->solvePoisson(*(SolvePoissonJob*)jobs[conf->threads*(2+i)+threadId]);
			barrier->wait();
		}
		
		// Project velocity field into divergence free field
		windField->project(*(ProjectJob*)jobs[conf->threads*(2+conf->sorIter)+threadId], timer.ddelta());
		barrier->wait();

		// Update snowflake positions
		fallingSnow->updateSnowflakes(*(SnowflakeJob*)jobs[conf->threads*(3+conf->sorIter)+threadId], min(timer.ddelta(), 0.05));
		barrier->wait();

		// Advance snowflake positions
		fallingSnow->advance(*(AdvanceSnowflakesJob*)jobs[conf->threads*(4+conf->sorIter)+threadId]);
		barrier->wait();

		// Create vertex array for snowcover
		scene->genVertexArray(*(SnowcoverJob*)jobs[conf->threads*(5+conf->sorIter)+threadId]);
		barrier->wait();

		if (threadId == 0)
		{
			renderJob();
			incrementTime();
		} 
		barrier->wait();
	}
}

void
SnowSimulator::mainLoopSerial()
{
	while (!conf->state.interrupted)
	{
		// Self-advect velocities
		windField->advect(*(AdvectJob*)jobs[0], timer.ddelta());

		// Create right-hand side of Ax = b
		windField->buildSolutionVector(*(BuildSolutionVectorJob*)jobs[1], timer.ddelta());

		// Solve Poisson equation using SOR
		for (int i = 0; i < conf->sorIter; i++)
		{
			windField->solvePoisson(*(SolvePoissonJob*)jobs[2+i]);
		}
		
		// Project velocity field into divergence free field
		windField->project(*(ProjectJob*)jobs[2+conf->sorIter], timer.ddelta());

		// Update snowflake positions
		fallingSnow->updateSnowflakes(*(SnowflakeJob*)jobs[3+conf->sorIter], min(timer.ddelta(), 0.05));
		fallingSnow->advance(*(AdvanceSnowflakesJob*)jobs[4+conf->sorIter]);

		// Create vertex array for snowcover
		scene->genVertexArray(*(SnowcoverJob*)jobs[5+conf->sorIter]);
		
		// Render
		renderJob();
		incrementTime();
	}

}

void
SnowSimulator::doJob(Job *job)
{
	switch (job->type)
	{
		case JOB_ADVECT:
			windField->advect(*(AdvectJob*) job, timer.ddelta());
			break;

		case JOB_BUILD_SOLUTION_VECTOR:
			windField->buildSolutionVector(*(BuildSolutionVectorJob*)job, timer.ddelta());
			break;

		case JOB_SOLVE_POISSON:
			windField->solvePoisson(*(SolvePoissonJob*)job);
			break;

		case JOB_PROJECT:
			windField->project(*(ProjectJob*)job, timer.ddelta());
			break;

		case JOB_SNOWFLAKE:
			fallingSnow->updateSnowflakes(*(SnowflakeJob*)job, min(timer.ddelta(), 0.05));
			break;

		case JOB_ADVANCE_SNOWFLAKES:
			fallingSnow->advance(*(AdvanceSnowflakesJob*)job);
			break;

		case JOB_SNOWCOVER:
			scene->genVertexArray(*(SnowcoverJob*)job);
			break;

		case JOB_RENDER:
			renderJob();
			break;

		case JOB_PREPARE_TIMESTEP:
			incrementTime();
			break;
	}
}

void
SnowSimulator::renderJob()
{
	// Handle user interaction
	renderer->processSDLEvents();

	// Render scene
	if (conf->state.render)
		renderer->render();

}

void 
SnowSimulator::incrementTime()
{
	// Increment time
	timer.increment();

	count.add(timer.idelta());
	frames++;

	// Print framerate
	printFps();

	/*printMutex.lock();
	std::cout << "Finished render job" << std::endl;
	printMutex.unlock();*/
}

void
SnowSimulator::printFps()
{
	if (timer.time() > nextPrintTime)
	{
		std::cout << count.getFramerate()  << " fps" <<  std::endl;
		nextPrintTime += conf->printFpsInterval;
	}
}

void
SnowSimulator::initJobs()
{
	// thr * ( advect + sol + x*poisson + project + snowflake +adv.snowflakes + snowcover ) + render + prepare
	//  = thr * (6+sorIter) + 2
	nJobs = conf->threads * (6 + conf->sorIter) + 2;
	jobs = new Job*[nJobs];
	
	// Set up self-advection
	for (int i = 0; i < conf->threads; i++)
	{
		int n = i;
		jobs[n] = new AdvectJob(*conf, i);
		jobs[n]->state = JOB_STATE_READY;
		jobs[n]->nParents = 1;
		jobs[n]->nParentsDone = 0;
		jobs[n]->nChildren = conf->threads;
		jobs[n]->children = new Job*[conf->threads];
	}

	// Set up build solution vector
	for (int i = 0; i < conf->threads; i++)
	{
		int n = conf->threads + i;
		jobs[n] = new BuildSolutionVectorJob(*conf, i);
		jobs[n]->state = JOB_STATE_BLOCKED;
		jobs[n]->nParents = conf->threads;
		jobs[n]->nParentsDone = 0;
		jobs[n]->nChildren = conf->threads;
		jobs[n]->children = new Job*[conf->threads];
	}

	// Set up SOR poisson solver
	for (int k = 0; k < conf->sorIter; k++)
	{
		for (int i = 0; i < conf->threads; i++)
		{
			int n = conf->threads * (2 + k)  + i;
			jobs[n] = new SolvePoissonJob(*conf, i);
			jobs[n]->state = JOB_STATE_BLOCKED;
			jobs[n]->nParents = conf->threads;
			jobs[n]->nParentsDone = 0;
			jobs[n]->nChildren = conf->threads;
			jobs[n]->children = new Job*[conf->threads];
		}
	}

	// Set up projection 
	for (int i = 0; i < conf->threads; i++)
	{
		int n = conf->threads * (2 + conf->sorIter) + i;
		jobs[n] = new ProjectJob(*conf, i);
		jobs[n]->state = JOB_STATE_BLOCKED;
		jobs[n]->nParents = conf->threads;
		jobs[n]->nParentsDone = 0;
		jobs[n]->nChildren = conf->threads;
		jobs[n]->children = new Job*[conf->threads];
	}

	// Set up snowflake movement
	for (int i = 0; i < conf->threads; i++)
	{
		int n = conf->threads * (3 + conf->sorIter) + i;
		jobs[n] = new SnowflakeJob(*conf, i);
		jobs[n]->state = JOB_STATE_BLOCKED;
		jobs[n]->nParents = conf->threads;
		jobs[n]->nParentsDone = 0;
		jobs[n]->nChildren = 1 + conf->threads;
		jobs[n]->children = new Job*[jobs[n]->nChildren];
	}
	
	// Set up advance snowflakes
	for (int i = 0; i < conf->threads; i++)
	{
		int n = conf->threads * (4 + conf->sorIter) + i;
		jobs[n] = new AdvanceSnowflakesJob(*conf, i);
		jobs[n]->state = JOB_STATE_BLOCKED;
		jobs[n]->nParents = 2; //conf->threads + 1;
		jobs[n]->nParentsDone = 1;
		jobs[n]->nChildren = 2;
		jobs[n]->children = new Job*[jobs[n]->nChildren];
	}
	// Set up snow cover generation
	for (int i = 0; i < conf->threads; i++)
	{
		int n = conf->threads * (5 + conf->sorIter) + i;
		jobs[n] = new SnowcoverJob(*conf, i);
		jobs[n]->state = JOB_STATE_BLOCKED;
		jobs[n]->nParents = conf->threads + 1;
		jobs[n]->nParentsDone = 1;
		jobs[n]->nChildren = 2;
		jobs[n]->children = new Job*[jobs[n]->nChildren];
	}
	
	// Set up rendering
	jobs[nJobs - 2] = new RenderJob(*conf);
	jobs[nJobs - 2]->state = JOB_STATE_BLOCKED;
	jobs[nJobs - 2]->nParents = conf->threads * 2;
	jobs[nJobs - 2]->nParentsDone = 0;
	jobs[nJobs - 2]->nChildren = conf->threads * 2;
	jobs[nJobs - 2]->children = new Job*[jobs[nJobs-2]->nChildren];
	jobs[nJobs - 2]->forcedThread = 0;

	// Set up preparation of next time step
	jobs[nJobs - 1] = new PrepareTimestepJob(*conf);
	jobs[nJobs - 1]->state = JOB_STATE_BLOCKED;
	jobs[nJobs - 1]->nParents = conf->threads * 2;
	jobs[nJobs - 1]->nParentsDone = 0;
	jobs[nJobs - 1]->nChildren = conf->threads;
	jobs[nJobs - 1]->children = new Job*[conf->threads];



	//// "FORWARD" DEPENDENCIES
	
	// Advect-job
	for (int i = 0; i < conf->threads; i++)
	{
		for (int j = 0; j < conf->threads; j++)
		{
			jobs[i]->children[j] = jobs[conf->threads + j];
		}
		jobs[i]->preferredSuccessor = jobs[conf->threads + i];
	}

	// Build solution vector job
	for (int i = 0; i < conf->threads; i++)
	{
		for (int j = 0; j < conf->threads; j++)
		{
			jobs[i + conf->threads]->children[j] = jobs[conf->threads * 2 + j];
		}
		jobs[i + conf->threads]->preferredSuccessor = jobs[conf->threads * 2 + i];
	}

	// SOR Poisson solver job
	for (int k = 0; k < conf->sorIter; k++)
	{
		for (int i = 0; i < conf->threads; i++)
		{
			for (int j = 0; j < conf->threads; j++)
			{
				jobs[i + conf->threads * (2+k)]->children[j] = jobs[conf->threads * (3+k) + j];
			}
			jobs[i + conf->threads * (2+k)]->preferredSuccessor = jobs[conf->threads * (3+k) + i];
		}
	}
	
	// Projection job
	for (int i = 0; i < conf->threads; i++)
	{
		for (int j = 0; j < conf->threads; j++)
		{
			jobs[i + conf->threads * (2+conf->sorIter)]->children[j] = jobs[conf->threads * (3+conf->sorIter) + j];
		}

		jobs[i + conf->threads * (2+conf->sorIter)]->preferredSuccessor = jobs[conf->threads * (3+conf->sorIter) + i];
	}
	
	// Snowflake job
	for (int i = 0; i < conf->threads; i++)
	{
		jobs[i + conf->threads * (3+conf->sorIter)]->children[0] = jobs[conf->threads * (4+conf->sorIter) + i];
		for (int j = 0; j < conf->threads; j++)
		{
			jobs[i + conf->threads * (3+conf->sorIter)]->children[1+j] = jobs[conf->threads * (5+conf->sorIter) + j];
		}
		jobs[i + conf->threads * (3+conf->sorIter)]->preferredSuccessor = jobs[conf->threads * (5+conf->sorIter) + i];
	}

	// Advance snowflakes job
	for (int i = 0; i < conf->threads; i++)
	{
		jobs[i + conf->threads * (4+conf->sorIter)]->children[0] = jobs[nJobs-2];
		jobs[i + conf->threads * (4+conf->sorIter)]->children[1] = jobs[nJobs-1];
		jobs[i + conf->threads * (4+conf->sorIter)]->preferredSuccessor = jobs[conf->threads * (5+conf->sorIter) + i];
	}
	// Snowcover job
	for (int i = 0; i < conf->threads; i++)
	{
		jobs[i + conf->threads * (5+conf->sorIter)]->children[0] = jobs[nJobs-2];
		jobs[i + conf->threads * (5+conf->sorIter)]->children[1] = jobs[nJobs-1];
		jobs[i + conf->threads * (5+conf->sorIter)]->preferredSuccessor = jobs[nJobs-2];
	}
	jobs[conf->threads * (5+conf->sorIter)]->preferredSuccessor = jobs[nJobs-2];

	// Render job
	for (int j = 0; j < conf->threads; j++)
	{
		jobs[nJobs-2]->children[j] = jobs[conf->threads * (4+conf->sorIter) + j];
		jobs[nJobs-2]->children[j+conf->threads] = jobs[conf->threads * (5+conf->sorIter) + j];
	}
	jobs[nJobs-2]->preferredSuccessor = jobs[0];
	
	// Prepare timestep job
	for (int j = 0; j < conf->threads; j++)
	{
		jobs[nJobs-1]->children[j] = jobs[j];
	}
	jobs[nJobs-1]->preferredSuccessor = jobs[0];


	/*
	for (int i = 0; i < nJobs; i++)
	{
		Job *job = jobs[i];
		job->print();
		cout << "   Children: ------------" << endl;
		for (int j = 0; j < job->nChildren; j++)
		{
			job->children[j]->print();
		}
		cout << "   ----------------------" << endl;
	}*/
	
}

SimulationThread::SimulationThread(int threadId, SnowSimulator *sim)
	: threadId(threadId), sim(sim)
{
}

void
SimulationThread::run()
{
	std::cout << "Thread " << threadId << " is running..." << std::endl;
	if (sim->conf->parallelMode == PARALLEL_DATA)
		sim->mainLoopDataParallel(threadId);
	else if (sim->conf->parallelMode == PARALLEL_PIPELINE)
		sim->mainLoopPipelined(threadId);
	else
		sim->mainLoopSerial();
}

