#include "snowflake.h"

const double Snowflake::g = 9.80665;

void
Snowflake::initialize(Configuration& conf)
{
	// 1. initialize diameter and density
	if (conf.temperature <= -0.061)
		diameter = 0.015f * powf(fabs((float)conf.temperature), -0.35f);
	else
		diameter = 0.04f;

	diameter *= float(rand() / ((double) RAND_MAX) + 0.5f); // vary +-50%

	if (conf.temperature <= -1) // dry
		rho = 0.170f / diameter;
	else                        // wet
		rho = 0.724f / diameter;

	// 2. initialize rotational radius, position and angular speed
	rotRadius = float(rand() / ((double) RAND_MAX) * 0.2f);
	
	spawn(conf.xSize, conf.ySize, conf.zSize);
	
	rotOmega = float(rand() / ((double) RAND_MAX) * (M_PI / 3 - M_PI / 4) + M_PI / 4);
	if (rand() < RAND_MAX / 2) rotOmega *= -1;

	// 3. initialize Vmaxz
	if (conf.temperature <= -1)
		Vmaxz = float(rand() / ((double) RAND_MAX) + 0.5f);
	else
		Vmaxz = float(rand() / ((double) RAND_MAX) + 1.0f);

	// 4. calculate m and Fgravity 
	m = 1.0f / 6.0f * M_PI * rho * diameter * diameter * diameter;
	Fgravity.z =  -m * this->g;
	Fgravity.x = Fgravity.y = 0.0f;

	rotTheta = float(rand() / ((double) RAND_MAX) * 2 * M_PI);
	V = Vector3D(0.0f, 0.0f, -Vmaxz);
	
	// copy into "twin"
	p_prev = p;
}

void
Snowflake::takeStep(double dt, WindField& wf)
{
	Vector3D Vwind;
	Vector3D Vcirc;
	Vector3D Vfluid;
	Vector3D Fdrag;
	Vector3D a;

	wf.getVelocity(&p_prev, &Vwind);
	//std::cout << "Wind velocity: " << Vwind.x << "," << Vwind.y << "," << Vwind.z << std::endl;
	
	float VwindAbs = sqrtf(Vwind.x*Vwind.x + Vwind.y*Vwind.y * Vwind.z*Vwind.z);
	float Vabs = sqrtf(V.x*V.x + V.y*V.y + V.z*V.z);
	//std::cout << "V: " << V.x << ", " << V.y << ", " << V.z << std::endl;

	// 5. calculate Vcirc(t)
	float Cvel = fabs(VwindAbs - Vabs) / Vabs;
	//std::cout << "Cvel: " << Cvel << std::endl;
	//Vector3D Vcirc = Vector3D(-sin(rotTheta), 0, cos(rotTheta)) * (Cvel * rotOmega * rotRadius);
	Vcirc.x = -sin(rotTheta)* Cvel * rotOmega * rotRadius;
	Vcirc.y = cos(rotTheta)* Cvel * rotOmega * rotRadius;
	Vcirc.z = 0.0f;
	

	// 6. calculate Vfluid and Fdrag
	//Vector3D Vfluid = Vwind - V;
	Vfluid.x = Vwind.x - V.x;
	Vfluid.y = Vwind.y - V.y;
	Vfluid.z = Vwind.z - V.z;
	float len2 = Vfluid.x*Vfluid.x + Vfluid.y*Vfluid.y + Vfluid.z*Vfluid.z;
	float ilen = 1.0f / sqrtf(len2);
	Vfluid.x *= ilen;
	Vfluid.y *= ilen;
	Vfluid.z *= ilen;
	//Vector3D Fdrag = Vfluid.unit() * (Vfluid.x*Vfluid.x + Vfluid.y*Vfluid.y + Vfluid.z*Vfluid.z) 
	//			* m * g / (Vmaxz * Vmaxz);
	Fdrag.x = Vfluid.x * len2 * m * g / (Vmaxz * Vmaxz);
	Fdrag.y = Vfluid.y * len2 * m * g / (Vmaxz * Vmaxz);
	Fdrag.z = Vfluid.z * len2 * m * g / (Vmaxz * Vmaxz);

	// 7. calculate a
	//cout << "m=" << m << ", m_prev=" << m_prev << ", rho=" << rho << ", diam=" << diameter << endl;
	//Vector3D a = (Fgravity + Fdrag) / m;
	a.x = (Fgravity.x + Fdrag.x) / m;
	a.y = (Fgravity.y + Fdrag.y) / m;
	a.z = (Fgravity.z + Fdrag.z) / m;
	
	// 8. calculate new position
	//p = p_prev + (V + Vcirc) * dt + a * 0.5 * dt * dt;
	p.x = p_prev.x + (V.x + Vcirc.x) * dt + a.x * 0.5 * dt * dt;
	p.y = p_prev.y + (V.y + Vcirc.y) * dt + a.y * 0.5 * dt * dt;
	p.z = p_prev.z + (V.z + Vcirc.z) * dt + a.z * 0.5 * dt * dt;

	// 9. calculate new velocity
	//V += a * dt;
	V.x += a.x * dt;
	V.y += a.y * dt;
	V.z += a.z * dt;

	// 10. update rotational position
	rotTheta += rotOmega * dt;
	//if (rotTheta > 2*M_PI) rotTheta -= 2*M_PI;
	
}

void
Snowflake::advance()
{
	p_prev.x = p.x;
	p_prev.y = p.y;
	p_prev.z = p.z;
}

void
Snowflake::respawn_x(float xPlane, float yMax, float zMax)
{
	p.x = xPlane;
	p.y = ((double) rand() / RAND_MAX) * yMax;
	p.z = ((double) rand() / RAND_MAX) * zMax;
	advance();
	V = Vector3D(0, 0, -Vmaxz);
}

void
Snowflake::respawn_y(float yPlane, float xMax, float zMax)
{
	p.y = yPlane;
	p.x = ((double) rand() / RAND_MAX) * xMax;
	p.z = ((double) rand() / RAND_MAX) * zMax;
	advance();
	V = Vector3D(0, 0, -Vmaxz);
}

void
Snowflake::respawn_z(float zPlane, float xMax, float yMax)
{
	p.z = zPlane;
	p.x = ((double) rand() / RAND_MAX) * xMax;
	p.y = ((double) rand() / RAND_MAX) * yMax;
	advance();
	V = Vector3D(0, 0, -Vmaxz);
}

void
Snowflake::spawn(float xMax, float yMax, float zMax)
{
	p.x = ((double) rand() / RAND_MAX) * xMax;
	p.y = ((double) rand() / RAND_MAX) * yMax;
	p.z = ((double) rand() / RAND_MAX) * zMax;
	advance();
	V = Vector3D(0, 0, -Vmaxz);
}

void
PaperSnowflake::update(double dt)
{
	double k1[4], k2[4], k3[4], k4[4];

	k1[0] = dt * du(_u, _v, _omega, _theta);
	k1[1] = dt * dv(_u, _v, _omega, _theta);
	k1[2] = dt * dOmega(_u, _v, _omega, _theta);
	k1[3] = dt * dTheta(_u, _v, _omega, _theta);

	k2[0] = dt * du(_u + 0.5*k1[0], _v + 0.5*k1[1], 
					   _omega + 0.5*k1[2], _theta * 0.5*k1[3]);
	k2[1] = dt * dv(_u + 0.5*k1[0], _v + 0.5*k1[1], 
					   _omega + 0.5*k1[2], _theta * 0.5*k1[3]);
	k2[2] = dt * dOmega(_u + 0.5*k1[0], _v + 0.5*k1[1], 
					   _omega + 0.5*k1[2], _theta * 0.5*k1[3]);
	k2[3] = dt * dTheta(_u + 0.5*k1[0], _v + 0.5*k1[1], 
					   _omega + 0.5*k1[2], _theta * 0.5*k1[3]);

	k3[0] = dt * du(_u + 0.5*k2[0], _v + 0.5*k2[1], 
					   _omega + 0.5*k2[2], _theta * 0.5*k2[3]);
	k3[1] = dt * dv(_u + 0.5*k2[0], _v + 0.5*k2[1], 
					   _omega + 0.5*k2[2], _theta * 0.5*k2[3]);
	k3[2] = dt * dOmega(_u + 0.5*k2[0], _v + 0.5*k2[1], 
					   _omega + 0.5*k2[2], _theta * 0.5*k2[3]);
	k3[3] = dt * dTheta(_u + 0.5*k2[0], _v + 0.5*k2[1], 
					   _omega + 0.5*k2[2], _theta * 0.5*k2[3]);


	k4[0] = dt * du(_u + k3[0], _v + k3[1], _omega + k3[2], _theta + k3[3]);
	k4[1] = dt * dv(_u + k3[0], _v + k3[1], _omega + k3[2], _theta + k3[3]);
	k4[2] = dt * dOmega(_u + k3[0], _v + k3[1], _omega + k3[2], _theta + k3[3]);
	k4[3] = dt * dTheta(_u + k3[0], _v + k3[1], _omega + k3[2], _theta + k3[3]);

	_u += (k1[0] + 2*k2[0] + 2*k3[0] + k4[0]) / 6.0;
	_v += (k1[1] + 2*k2[1] + 2*k3[1] + k4[1]) / 6.0;
	_omega += (k1[2] + 2*k2[2] + 2*k3[2] + k4[2]) / 6.0;
	_theta += (k1[3] + 2*k2[3] + 2*k3[3] + k4[3]) / 6.0;
}

double 
PaperSnowflake::du(double u, double v, double omega, double theta)
{
	double V2 = u*u + v*v;
	double cosAlpha = -v / sqrt(V2);
	double alpha = atan(u/v);
	double cosTheta = cos(theta);
	double sinTheta = sin(theta);
	int sign = 1;
	
	if ( (v < 0 && ((alpha + theta) > 0) && ((alpha + theta) <  M_PI)) ||
	     (v > 0 && ((alpha + theta) < 0) && ((alpha + theta) > -M_PI)) )
		sign = -1;

	return (_kPerp - _kPara) * sinTheta * cosTheta * v
		- (_kPerp * sinTheta*sinTheta + _kPara * cosTheta*cosTheta) * u
		+ sign * M_PI * _rho * V2 * cos(alpha + theta) * cosAlpha;
}

double
PaperSnowflake::dv(double u, double v, double omega, double theta)
{
	double V2 = u*u + v*v;
	double alpha = atan(u/v); 
	double cosTheta = cos(theta);
	double sinTheta = sin(theta);
	int sign = -1;
	
	if ( (v < 0 && ((alpha + theta) > 0) && ((alpha + theta) <  M_PI)) ||
	     (v > 0 && ((alpha + theta) < 0) && ((alpha + theta) > -M_PI)) )
		sign = 1;

	return (_kPerp - _kPara) * sinTheta * cosTheta * u
		- (_kPerp * cosTheta*cosTheta + _kPara * sinTheta*sinTheta) * v
		+ sign * M_PI * _rho * V2 * cos(alpha + theta) * sin(alpha) - g;
		
}

double
PaperSnowflake::dOmega(double u, double v, double omega, double theta)
{
	double V2 = u*u + v*v;
	double alpha = atan(u/v);
	
	return -_kPerp * omega 
		- (3 * M_PI * _rho * V2 / _l) 
		* cos(alpha + theta) * sin(alpha + theta);
}
		
double
PaperSnowflake::dTheta(double u, double v, double omega, double theta)
{
	return omega;
}
	
