#include "wind_field.h"

using namespace std;



WindField::WindField(Configuration *conf, SceneDescriptor *scene)
	: conf(conf), scene(scene)
{
	Nx = conf->Nx;
	Ny = conf->Ny;
	Nz = conf->Nz;
	xSize = conf->wfSize.x;
	ySize = conf->wfSize.y;
	zSize = conf->wfSize.z;
	N = (Nx+2)*(Ny+2)*(Nz+2);
	float hx = xSize / Nx;
	float hy = ySize / Ny;
	float hz = zSize / Nz;
	if (hx != hy || hx != hz || hy != hz)
	{
		cout << "Error: Voxels in wind field are not equal-sized" << endl;
		exit(1);
	}

	h = 1.0f; // Using voxelsize = 1, and scaling world coordinates instread.
	
	initialize();
}

WindField::~WindField()
{
	delete [] m_obstacles;
	delete [] m_p;
	delete [] m_zPrev;
	delete [] m_yPrev;
	delete [] m_xPrev;
	delete [] m_zStar;
	delete [] m_yStar;
	delete [] m_xStar;
	delete [] m_z;
	delete [] m_y;
	delete [] m_x;
	delete [] m_b;
}

void
WindField::initialize()
{
	m_b = new float[N];
	m_x = new float[N];
	m_y = new float[N];
	m_z = new float[N];
	m_xStar = new float[N];
	m_yStar = new float[N];
	m_zStar = new float[N];
	m_xPrev = new float[N];
	m_yPrev = new float[N];
	m_zPrev = new float[N];
	m_p = new float[N];
	m_obstacles = new int[N];

	for (int i = 0; i < N; i++)
	{
		m_obstacles[i] = 0;
	}

	reset();
	buildObstacles();

}

void
WindField::reset()
{
	for (int i = 0; i < N; i++)
	{
		m_b[i] = m_x[i] = m_y[i] = m_z[i] = m_xStar[i] = m_yStar[i] = m_zStar[i]
				= m_xPrev[i] = m_yPrev[i] = m_zPrev[i] = m_p[i] = 0.0f;
	}

	for (int k = 1; k <= Nz; k++)
		for (int j = 1; j <= Ny; j++)
			for (int i = 1; i <= Nx; i++)
			{
				m_x[I(i,j,k)] = conf->state.boundaryVelocity.x;
				m_y[I(i,j,k)] = conf->state.boundaryVelocity.y;
				m_z[I(i,j,k)] = conf->state.boundaryVelocity.z;
			}
}

void
WindField::addForces(double dt)
{
	int k, j, i;
	for (k = 5; k < 7; k++)
		for (j = 5; j < 7; j++)
			for (i = 5; i < 7; i++)
			{
				//m_z[I(i, j, k)] += dt * -1.0;
			}
}

void
WindField::advect(AdvectJob& job, double dt)
{
	int i, j, k;
	int i0, j0, k0, i1, j1, k1;
	float x, y, z, s0, t0, u0, s1, t1, u1;
	float lower, upper;

	if (conf->state.resetWindField)
	{
		conf->state.resetWindField = false;
		reset();
	}
	
	for (k = job.start + 1; k <= job.end + 1; k++)
		for (j = 1; j <= Ny; j++)
			for (i = 1; i <= Nx; i++)
			{
				x = i - dt * m_x[I(i, j, k)];
                y = j - dt * m_y[I(i, j, k)];
				z = k - dt * m_z[I(i, j, k)];

				// if outside of domain, truncate values
				x = max(x, 0.5f);
				x = min(x, Nx + 0.5f);
				y = max(y, 0.5f);
				y = min(y, Ny + 0.5f);
				z = max(z, 0.5f);
				z = min(z, Nz + 0.5f);

				i0 = (int) x;
				i1 = i0 + 1;
				j0 = (int) y;
				j1 = j0 + 1;
				k0 = (int) z;
				k1 = k0 + 1;
                
				s1 = x - i0;
                s0 = 1 - s1;
                t1 = y - j0;
                t0 = 1 - t1;
				u1 = z - k0;
				u0 = 1 - u1;

				// interpolate x
				lower = s0 * (t0 * m_x[I(i0, j0, k0)] + t1 * m_x[I(i0, j1, k0)])
                                 + s1 * (t0 * m_x[I(i1, j0, k0)] + t1 * m_x[I(i1, j1, k0)]);

				upper = s0 * (t0 * m_x[I(i0, j0, k1)] + t1 * m_x[I(i0, j1, k1)])
                                 + s1 * (t0 * m_x[I(i1, j0, k1)] + t1 * m_x[I(i1, j1, k1)]);

				m_xStar[I(i,j,k)] = u0 * lower + u1 * upper;

				// interpolate y
				lower = s0 * (t0 * m_y[I(i0, j0, k0)] + t1 * m_y[I(i0, j1, k0)])
                                 + s1 * (t0 * m_y[I(i1, j0, k0)] + t1 * m_y[I(i1, j1, k0)]);

				upper = s0 * (t0 * m_y[I(i0, j0, k1)] + t1 * m_y[I(i0, j1, k1)])
                                 + s1 * (t0 * m_y[I(i1, j0, k1)] + t1 * m_y[I(i1, j1, k1)]);

				m_yStar[I(i,j,k)] = u0 * lower + u1 * upper;

				// interpolate z
				lower = s0 * (t0 * m_z[I(i0, j0, k0)] + t1 * m_z[I(i0, j1, k0)])
                                 + s1 * (t0 * m_z[I(i1, j0, k0)] + t1 * m_z[I(i1, j1, k0)]);

				upper = s0 * (t0 * m_z[I(i0, j0, k1)] + t1 * m_z[I(i0, j1, k1)])
                                 + s1 * (t0 * m_z[I(i1, j0, k1)] + t1 * m_z[I(i1, j1, k1)]);

				m_zStar[I(i,j,k)] = u0 * lower + u1 * upper;
			}

	setVelocityBoundaries(m_xStar, m_yStar, m_zStar, job.start, job.end, job.bigZStart, job.bigZEnd, job.bigYStart, job.bigYEnd);
}

void
WindField::buildSolutionVector(BuildSolutionVectorJob& job, double dt)
{
	int i, j, k;
	float h_inv = 0.5f / (h * dt);

	for (k = 1+job.start; k <= 1+job.end; k++)
		for (j = 1; j <= Ny; j++)
			for (i = 1; i <= Nx; i++)
			{
				m_b[I(i, j, k)] = h_inv * ((m_obstacles[I(i,j,k)]^1)&1) * (
								(m_xStar[I(i+1, j, k)] - m_xStar[I(i-1, j, k)]) +
								(m_yStar[I(i, j+1, k)] - m_yStar[I(i, j-1, k)]) +
								(m_zStar[I(i, j, k+1)] - m_zStar[I(i, j, k-1)]));
				m_p[I(i,j,k)] = 0.0f;
			}

	setBoundaries(m_b, job.start, job.end, job.bigZStart, job.bigZEnd, job.bigYStart, job.bigYEnd);
	setBoundaries(m_p, job.start, job.end, job.bigZStart, job.bigZEnd, job.bigYStart, job.bigYEnd);
}

void
WindField::solvePoisson(SolvePoissonJob& job)
{
	int i, j, k;
	float w = conf->sorRelax;
	float temp;
	int div;

	for (k = 1+job.start; k <= job.end; k++)
	{
		for (j = 1; j <= Ny; j++)
		{
			for (i = 1; i <= Nx; i++)
			{
				temp = 0.0f;
				div = 0;
				div += (m_obstacles[I(i-1,j,k)]^1) & 1;
				div += (m_obstacles[I(i+1,j,k)]^1) & 1;
				div += (m_obstacles[I(i,j-1,k)]^1) & 1;
				div += (m_obstacles[I(i,j+1,k)]^1) & 1;
				div += (m_obstacles[I(i,j,k-1)]^1) & 1;
				div += (m_obstacles[I(i,j,k+1)]^1) & 1;

				if (div > 0) 
				{
					m_p[I(i, j, k)] = (1-w) * m_p[I(i,j,k)] + w * ((
								m_p[I(i-1, j, k)] * (!m_obstacles[I(i-1,j,k)] & 1) + 
								m_p[I(i+1, j, k)] * (!m_obstacles[I(i+1,j,k)] & 1) +
								m_p[I(i, j-1, k)] * (!m_obstacles[I(i,j-1,k)] & 1) + 
								m_p[I(i, j+1, k)] * (!m_obstacles[I(i,j+1,k)] & 1) +
								m_p[I(i, j, k-1)] * (!m_obstacles[I(i,j,k-1)] & 1) + 
								m_p[I(i, j, k+1)] * (!m_obstacles[I(i,j,k+1)] & 1) 
								- m_b[I(i, j, k)]) / float(div)) * ((m_obstacles[I(i,j,k)]^1) & 1);
				}
				else
				{
					m_p[I(i, j, k)] = 0.0f;
				}
			}
		}
	}

}

void
WindField::project(ProjectJob& job, double dt)
{
	int i, j, k, index;
	float factor = 0.5f * dt / h;

	setBoundaries(m_p, job.start, job.end, job.bigZStart, job.bigZEnd, job.bigYStart, job.bigYEnd);

	for (k = 1+job.start; k <= job.end; k++)
		for (j = 1; j <= Ny; j++)
			for (i = 1; i <= Nx; i++)
			{
				index = I(i, j, k);

				// advance
				m_xPrev[index] = m_x[index];
				m_yPrev[index] = m_y[index];
				m_zPrev[index] = m_z[index];

				m_x[index] = m_xStar[index] - factor * (m_p[I(i+1, j, k)] - m_p[I(i-1, j, k)]);
                m_y[index] = m_yStar[index] - factor * (m_p[I(i, j+1, k)] - m_p[I(i, j-1, k)]);
                m_z[index] = m_zStar[index] - factor * (m_p[I(i, j, k+1)] - m_p[I(i, j, k-1)]);

			}

	setVelocityBoundaries(m_x, m_y, m_z, job.start, job.end, job.bigZStart, 
					job.bigZEnd, job.bigYStart, job.bigYEnd);
	setVelocityBoundaries(m_xPrev, m_yPrev, m_zPrev, job.start, job.end, job.bigZStart, 
					job.bigZEnd, job.bigYStart, job.bigYEnd);
}



void
WindField::setBoundaries(	float *f, 
							int zFrom, /* indices for parallelization */
							int zTo,
							int bigZFrom,
							int bigZTo, 
							int bigYFrom, 
							int bigYTo)
{
	int i, j, k, index, mask;

	for (k = bigZFrom; k < bigZTo; k++)
		for (j = 0; j <= Ny+1; j++)
			{
				f[I(0, j, k)] = f[I(1, j, k)];
				f[I(Nx+1, j, k)] = f[I(Nx, j, k)];
			}	

	for (k = bigZFrom; k < bigZTo; k++)
		for (i = 0; i <= Nx+1; i++)
			{
				f[I(i, 0, k)] = f[I(i, 1, k)];
				f[I(i, Ny+1, k)] = f[I(i, Ny, k)];
			}

	for (j = bigYFrom; j < bigYTo; j++)
		for (i = 0; i <= Nx+1; i++)
			{
				f[I(i, j, 0)] = f[I(i, j, 1)];
				f[I(i, j, Nz+1)] = f[I(i, j, Nz)];
			}
	
	for (k = 1+zFrom; k <= zTo; k++)
		for (j = 1; j <= Ny; j++)
			for (i = 1; i <= Nx; i++)
			{
				index = I(i, j, k);
				mask = m_obstacles[index];

				if (mask & VOXEL_MASK_SELF)
				{
					if (!(mask & VOXEL_MASK_LEFT))
					{
						f[index] = f[I(i-1,j,k)];
					}
					else if (!(mask & VOXEL_MASK_RIGHT))
					{
						f[index] = f[I(i+1,j,k)];
					}
					else if (!(mask & VOXEL_MASK_TOP))
					{
						f[index] = f[I(i, j, k+1)];
					}
					else if (!(mask & VOXEL_MASK_BOTTOM))
					{
						f[index] = f[I(i, j, k-1)];
					}
					else if (!(mask & VOXEL_MASK_FRONT))
					{
						f[index] = f[I(i, j-1, k)];
					}
					else if (!(mask & VOXEL_MASK_BACK))
					{
						f[index] = f[I(i, j+1, k)];
					}
					else
					{
						f[index] = 0.0;
					}
				}
			}
}


void
WindField::setVelocityBoundaries(	float *x, 
									float *y, 
									float *z, 
									int zFrom, /* indices for parallelization */
									int zTo,
									int bigZFrom,
									int bigZTo, 
									int bigYFrom, 
									int bigYTo)
{
	int i, j, k, index, mask;

	for (k = 1+zFrom; k <= zTo; k++)
		for (j = 1; j <= Ny; j++)
			for (i = 1; i <= Nx; i++)
			{
				index = I(i, j, k);
				mask = m_obstacles[index];
				if (mask & VOXEL_MASK_SELF)
				{
					if (!(mask & VOXEL_MASK_LEFT))
					{
						x[index] = 0.0;
						//x[index] = -x[I(i-1, j, k)];
						y[index] = y[I(i-1, j, k)];
						z[index] = z[I(i-1, j, k)];
					}
					else if (!(mask & VOXEL_MASK_RIGHT))
					{
						x[index] = 0.0;
						//x[index] = -x[I(i+1, j, k)];
						y[index] = y[I(i+1, j, k)];
						z[index] = z[I(i+1, j, k)];
					}
					else if (!(mask & VOXEL_MASK_TOP))
					{
						x[index] = x[I(i, j, k+1)];
						y[index] = y[I(i, j, k+1)];
						//z[index] = -z[I(i, j, k+1)];
						z[index] = 0.0;
					}
					else if (!(mask & VOXEL_MASK_BOTTOM))
					{
						x[index] = x[I(i, j, k-1)];
						y[index] = y[I(i, j, k-1)];
						//z[index] = -z[I(i, j, k-1)];
						z[index] = 0.0;
					}
					else if (!(mask & VOXEL_MASK_FRONT))
					{
						x[index] = x[I(i, j-1, k)];
						//y[index] = -y[I(i, j-1, k)];
						y[index] = 0.0;
						z[index] = z[I(i, j-1, k)];
					}
					else if (!(mask & VOXEL_MASK_BACK))
					{
						x[index] = x[I(i, j+1, k)];
						//y[index] = -y[I(i, j+1, k)];
						y[index] = 0.0;
						z[index] = z[I(i, j+1, k)];
					}
					else
					{
						x[index] = 0.0;
						y[index] = 0.0;
						z[index] = 0.0;
					}
				}
			}

	for (k = bigZFrom; k < bigZTo; k++)
		for (j = 0; j <= Ny+1; j++)
			{
				x[I(0, j, k)] = conf->state.boundaryVelocity.x;
				x[I(Nx+1, j, k)] = conf->state.boundaryVelocity.x;
				y[I(0, j, k)] = conf->state.boundaryVelocity.y;
				y[I(Nx+1, j, k)] = conf->state.boundaryVelocity.y;
				z[I(0, j, k)] = conf->state.boundaryVelocity.z;
				z[I(Nx+1, j, k)] = conf->state.boundaryVelocity.z;
			}	

	for (k = bigZFrom; k < bigZTo; k++)
		for (i = 0; i <= Nx+1; i++)
			{
				x[I(i, 0, k)] = conf->state.boundaryVelocity.x;
				x[I(i, Ny+1, k)] = conf->state.boundaryVelocity.x;
				y[I(i, 0, k)] = conf->state.boundaryVelocity.y;
				y[I(i, Ny+1, k)] = conf->state.boundaryVelocity.y;
				z[I(i, 0, k)] = conf->state.boundaryVelocity.z;
				z[I(i, Ny+1, k)] = conf->state.boundaryVelocity.z;
			}

	for (j = bigYFrom; j < bigYTo; j++)
		for (i = 0; i <= Nx+1; i++)
			{
				x[I(i, j, 0)] = x[I(i, j, 1)]; //conf->state.boundaryVelocity.x;
				x[I(i, j, Nz+1)] = conf->state.boundaryVelocity.x;
				y[I(i, j, 0)] = y[I(i, j, 1)]; //conf->state.boundaryVelocity.y;
				y[I(i, j, Nz+1)] = conf->state.boundaryVelocity.y;
				z[I(i, j, 0)] = 0.0;
				z[I(i, j, Nz+1)] = z[I(i, j, Nz)];
			}


}

/**
 * Interpolate velocity at pos. 
 */
void
WindField::getVelocity(Vector3D *pos, Vector3D *ret)
{
	int i0, j0, k0, i1, j1, k1;
	float x, y, z, s0, t0, u0, s1, t1, u1;
	float lower, upper;
	float vs = Nx / xSize;
	
	x = pos->x - conf->wfPos.x;
	y = pos->y - conf->wfPos.y;
	z = pos->z - conf->wfPos.z;

	if (z < 0.5f &&
		x > 0.5f &&
		x < conf->wfSize.x + 0.5f &&
		y > 0.5f &&
		y < conf->wfSize.y + 0.5f &&
		z < conf->wfSize.z + 0.5f)
	{
		z = 0.5f;
	}
	else if (x < 0.5f ||
		x > conf->wfSize.x + 0.5f ||
		y < 0.5f ||
		y > conf->wfSize.y + 0.5f ||
		z < 0.5f ||
		z > conf->wfSize.z + 0.5f)
	{
		ret->x = conf->state.boundaryVelocity.x;
		ret->y = conf->state.boundaryVelocity.y;
		ret->z = conf->state.boundaryVelocity.z;
		return;
	}

	i0 = (int) (x * vs);
	i1 = i0 + 1;
	j0 = (int) (y * vs);
	j1 = j0 + 1;
	k0 = (int) (z * vs);
	k1 = k0 + 1;

	s1 = x * vs- i0;
    s0 = 1 - s1;
    t1 = y * vs - j0;
    t0 = 1 - t1;
	u1 = z * vs - k0;
	u0 = 1 - u1;
	
	// interpolate x
	lower = s0 * (t0 * m_xPrev[I(i0, j0, k0)] + t1 * m_xPrev[I(i0, j1, k0)])
                 + s1 * (t0 * m_xPrev[I(i1, j0, k0)] + t1 * m_xPrev[I(i1, j1, k0)]);

	upper = s0 * (t0 * m_xPrev[I(i0, j0, k1)] + t1 * m_xPrev[I(i0, j1, k1)])
                  + s1 * (t0 * m_xPrev[I(i1, j0, k1)] + t1 * m_xPrev[I(i1, j1, k1)]);

	ret->x = u0 * lower + u1 * upper;

	// interpolate y
	lower = s0 * (t0 * m_yPrev[I(i0, j0, k0)] + t1 * m_yPrev[I(i0, j1, k0)])
                    + s1 * (t0 * m_yPrev[I(i1, j0, k0)] + t1 * m_yPrev[I(i1, j1, k0)]);

	upper = s0 * (t0 * m_yPrev[I(i0, j0, k1)] + t1 * m_yPrev[I(i0, j1, k1)])
                     + s1 * (t0 * m_yPrev[I(i1, j0, k1)] + t1 * m_yPrev[I(i1, j1, k1)]);

	ret->y = u0 * lower + u1 * upper;

	// interpolate z
	lower = s0 * (t0 * m_zPrev[I(i0, j0, k0)] + t1 * m_zPrev[I(i0, j1, k0)])
                     + s1 * (t0 * m_zPrev[I(i1, j0, k0)] + t1 * m_zPrev[I(i1, j1, k0)]);

	upper = s0 * (t0 * m_zPrev[I(i0, j0, k1)] + t1 * m_zPrev[I(i0, j1, k1)])
                      + s1 * (t0 * m_zPrev[I(i1, j0, k1)] + t1 * m_zPrev[I(i1, j1, k1)]);

	ret->z = u0 * lower + u1 * upper;
}

void
WindField::buildObstacles()
{
	float domainX, domainY, domainZ;
	float vs = xSize / Nx;

	for (int k = 1; k <= Nz; k++)
		for (int j = 1; j <= Ny; j++)
			for (int i = 1; i <= Nx; i++)
			{
				domainX = (i-0.5) * vs + conf->wfPos.x;
				domainY = (j-0.5) * vs + conf->wfPos.y;
				domainZ = (k-0.5) * vs + conf->wfPos.z;

				for (int ib = 0; ib < conf->nBoxes-1; ib++)
				{
					Box& b = scene->boxes[ib];

					if (domainX > b.pos.x && domainX < (b.pos.x + b.size.x) &&
						domainY > b.pos.y && domainY < (b.pos.y + b.size.y) &&
						domainZ > b.pos.z && domainZ < (b.pos.z + b.size.z) )
					{
						m_obstacles[I(i,j,k)] = VOXEL_MASK_SELF;
					}
				}
			}
					
	for (int k = 1; k <= Nz; k++)
		for (int j = 1; j <= Ny; j++)
			for (int i = 1; i <= Nx; i++)
			{
				m_obstacles[I(i,j,k)] += 
						(m_obstacles[I(i-1,j,k)] & 1) * VOXEL_MASK_LEFT +
						(m_obstacles[I(i+1,j,k)] & 1) * VOXEL_MASK_RIGHT +
						(m_obstacles[I(i,j-1,k)] & 1) * VOXEL_MASK_FRONT +
						(m_obstacles[I(i,j+1,k)] & 1) * VOXEL_MASK_BACK +
						(m_obstacles[I(i,j,k-1)] & 1) * VOXEL_MASK_BOTTOM +
						(m_obstacles[I(i,j,k+1)] & 1) * VOXEL_MASK_TOP;
			}
}
